(* gamutctlr.ml     Controller for Gamut example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 *)

let frnd f =
    floor (f +. 0.5)

let hsva_of_hsla (h, s, l, a) =
    (* Convert HSLA to HSVA color representation.
     *)
    let os =
        if l <= 0.5 then
            (2.0 *. s *. l) /. (l +. s *. l)
        else
            (2.0 *. s *. (1.0 -. l)) /. (l +. s *. (1.0 -. l))
    in let ov =
        if l <= 0.5 then
            l +. s *. l
        else
            l +. s *. (1.0 -. l)
    in
        (h, os, ov, a)


(* What kind of touch right now, based on where the touch *started*.
 *)
type touchtype =
    | TouchNone    (* No touch at the moment *)
    | TouchCircle  (* Touching the circle in the middle *)
    | TouchBack    (* Touching the background *)


class t =
object (self)
    inherit Wrappee.t

    val interval: float = 1.0 /. 24.0 (* Interval between timer ticks *)

    (* View that delegates to us.  If this fails to show up, check the
     * NIB file, Gamut.xib.
     *)
    val mutable theDelegator: UiView.t option = None

    (* fgtime is incremented automatically by timer ticks.  bgtime is a
     * record of the most recent synchronization of the background and
     * foreground times, but is otherwise static.  Times control
     * lightness of colors.
     *)
    val mutable fgtime: float = 0.0
    val mutable bgtime: float = 6.5

    (* State of touches.
     *)
    val mutable touchty: touchtype = TouchNone (* What kind of touch now *)
    val mutable touchpoint: Cocoa.point =      (* Most recent touched spot *)
        (0.0, 0.0)

    (* The circle in the middle of the view.
     *)
    val mutable theCircle : UiBezierPath.t option = None

    (* The text in the circle.
     *)
    val label = "\xce\xa8" (* Greek capital psi *)
    val mutable labelsize = (0.0, 0.0)
    val mutable theLabelfont : UiFont.t option = None

    (* Accessors.
     *)
    method delegator =
        match theDelegator with
        | None -> raise Not_found
        | Some d -> d

    method setDelegator' view =
        begin
        theDelegator <- Some view;
        let (_, _, w, h) = view#frame
        in let diam = frnd (w /. 3.0)
        in let rect = (diam, frnd ((h -. diam) /. 2.0), diam, diam)
        in let bzp = UiBezierPath.bezierPathWithOvalInRect' rect
        in let () = bzp#setLineWidth' (frnd (w /. 100.0))
        in
            begin
            theCircle <- Some bzp;
            touchpoint <- (w, 0.0694 *. h);
            end;
        end

    (* Application state change events.
     *)
    method applicationDidFinishLaunching' (app: UiApplication.t) =
        begin
        (* Startup code.
         *)
        ignore (
            NsTimer.scheduledTimer5 interval self "timerTick'" Wrappee.nil true
        );
        end

    method applicationDidReceiveMemoryWarning' (app: UiApplication.t) =
        Gc.compact ();  (* Best effort to reclaim space *)

    method applicationWillResignActive' (app: UiApplication.t) =
        ()

    method applicationDidBecomeActive' (app: UiApplication.t) =
        self#display

    method applicationWillTerminate' (app: UiApplication.t) =
        ()

    (* Touch events.
     *)
    method view'touchesBegan' (view: UiView.t) (touch: Cocoa.point) =
        begin
        (match theCircle with
        | None -> touchty <- TouchBack
        | Some c ->
            touchty <-
                if c#containsPoint' touch then TouchCircle else TouchBack
        );
        if touchty <> TouchCircle then
            touchpoint <- touch;
        end

    method view'touchesMoved' (view: UiView.t) (touch: Cocoa.point) =
        begin
        if touchty <> TouchCircle then
            touchpoint <- touch;
        end

    method view'touchesEnded' (view: UiView.t) (touch: Cocoa.point) =
        begin
        if touchty = TouchCircle then
            bgtime <- fgtime;
        touchty <- TouchNone;
        end

    method view'touchesCancelled' (view: UiView.t) (touch: Cocoa.point) =
        begin
        if touchty = TouchCircle then
            bgtime <- fgtime;
        touchty <- TouchNone;
        end


    (* Timer ticks.
     *)
    method timerTick' (timer: NsTimer.t) =
        begin
        fgtime <- fgtime +. interval;
        self#display;
        end


    (* Draw method.
     *)
    method view'drawRect' (v: UiView.t) (r: Cocoa.rect) =
        let white = (0.0, 0.0, 1.0, 1.0)
        in let kernfact = -0.1  (* Move to the left a little *)
        in let basefact = 0.08  (* Move down a little *)
        in let (_, _, vw, vh) = v#frame
        in let (fgcolor, _, resting) = self#colors_of_time fgtime
        in let bgtime' = if touchty = TouchCircle then fgtime else bgtime
        in let (bgcolor, bordcolor, _) = self#colors_of_time bgtime'
        in let labelfont = self#labelfont vw vh
        in let (lw, lh) = self#labelsize labelfont
        in 
            begin
            UiKit.set bgcolor;
            UiKit.rectFill (0.0, 0.0, vw, vh);
            (match theCircle with
            | None -> ()
            | Some c ->
                begin
                UiKit.set fgcolor;
                c#fill;
                UiKit.set (if resting then white else bordcolor);
                c#stroke;
                end
            );
            UiKit.set bgcolor;
            let pt =
                (frnd ((vw -. lw) /. 2.0) +. kernfact *. lw,
                 frnd ((vh -. lh) /. 2.0) +. basefact *. lw)
            in
                UiKit.string'drawAtPoint'withFont' label pt labelfont;
            end

    (* Private methods.
     *)
    method private display =
        match theDelegator with
        | None -> ()
        | Some v -> v#setNeedsDisplay

    method private colors_of_time time =
        (* Translate the current touch position and the given time into
         * an interesting color and a contrasting color and an indicator
         * of the resting period (when the color rests at base lightness
         * of 0.5).  We use HSLA internally, but return a color in HSVA
         * format.
         *)
        let (lightness, resting) = self#lightness_of_time time
        in let (ih, is, il, ia) as interior =
            match theDelegator with
            | None -> (0.0, 0.0, 0.0, 1.0)   (* Not really possible *)
            | Some view ->
                let (_, _, vw, vh) = view#frame
                in let (x, y) = touchpoint
                in let h = y /. vh
                in let s = x /. vw
                in let l = lightness
                in
                    (h, s, l, 1.0)
        in let border =
            if il > 0.5 then (ih, is, il -. 0.2, ia)
            else (ih, is, il +. 0.2, ia)
        in
            (hsva_of_hsla interior, hsva_of_hsla border, resting)

    method private lightness_of_time time =
        (* Return a lightness value (between 0.0 and 1.0) corresponding
         * to the given time, and a boolean saying whether we're in a
         * "rest" phase at the base lightness level (0.5).
         *
         * We want to alternate between light and dark with some fairly
         * longish period, with most of the time spent in the
         * interesting lightnesses between (say) 0.2 and 0.8.  There's
         * also a rest phase at the base level (0.5).
         *)
        let phases = [
            (2.0, 0.5); (* Rest phase is first *)
            (1.5, 0.8); (0.5, 1.0); (0.5, 0.8); (1.5, 0.5);
            (1.5, 0.2); (0.5, 0.0); (0.5, 0.2); (1.5, 0.5);
        ]
        in let total = List.fold_left (fun n (a, _) -> n +. a) 0.0 phases
        in let btime = mod_float time total
        in let rec go (prev, first) time phases =
            match phases with
            | [] -> (prev, first)
            | (seconds, level) :: _ when time <= seconds ->
                (prev +. (time /. seconds) *. (level -. prev), first)
            | (seconds, level) :: rest ->
                go (level, false) (time -. seconds) rest
        in
            go (0.5, true) btime phases

    method private labelfont vw vh =
        match theLabelfont with
        | Some f -> f
        | None ->
            let f = UiFont.fontWithName'size' "Helvetica-Oblique" (vw /. 4.0)
            in let () = theLabelfont <- Some f
            in
                f

    method private labelsize labelfont =
        match labelsize with
        | (0.0, 0.0) ->
            let sz = UiKit.string'sizeWithFont' label labelfont
            in let () = labelsize <- sz
            in
                sz
        | _ -> labelsize
end

let () =
    let wrapped robjcv =
        let c = new t in let () = c#setContainer robjcv in c
    in
        Callback.register "Gamutctlr.wrapped" wrapped
